
/*+*******************************************************************

 File:		DevSignal.h

 Project:	Deveice Servers with SUN-RPC

 Description:	Include file for a standard signal handling under
		HPUX, SUN and OS9.
		The definition of additional signal codes for OS9
		allows the use of the same makros on every system. 
		 
 Author(s):	Jens Meyer 

 Original:	June, 1991

 $Log:	DevSignal.h,v $
 * Revision 8.14  2000/06/22  14:50:41  14:50:41  goetz (Andrew GOETZ)
 * fixed bug in dev_event_listen() which caused server to crash
 * 
 Revision 8.13  2000/06/02 21:56:56  goetz
 _DEVICE_H now _TANGO_DEVICE_H in Device.H; (char*)taco_tango

 Revision 8.12  2000/06/02 15:58:13  goetz
 dev_event_fire() now device specific; ported to SuSE V6.4

 Revision 8.11  2000/05/31 07:48:01  goetz
 tango_api has local copies of argc and argv, ported to HP-UX

 * Revision 8.10  2000/05/29  21:39:26  21:39:26  goetz (Andrew GOETZ)
 * fixed prototyping problems with C++ on HP-UX
 * 
 * Revision 8.9  2000/05/29  18:13:36  18:13:36  goetz (Andrew GOETZ)
 * fixed bug in dev_import_timeout() which prevented timeouts < 2s
 * 
 Revision 8.8  2000/05/02 15:32:21  goetz
 added prototype for tsleep() for OS9 C++

 * Revision 8.7  2000/05/02  14:19:42  14:19:42  goetz (Andy Goetz)
 * removed check for device server running on different host
 * 
 * Revision 8.6  2000/05/02  13:29:35  13:29:35  goetz (Andy Goetz)
 * added random sleep to gettransient + 3 retries
 * 
 Revision 8.5  2000/03/27 17:13:48  goetz
 declared pmap_getmaps() as external C function for OS9 C++

 Revision 8.4  2000/03/13 14:50:05  goetz
 import timeout now programmable using dev_import_timeout()

 Revision 8.3  2000/03/13 10:31:38  goetz
 added Jens' modifications to fix bug in local putget

 Revision 8.2  2000/03/10 17:01:26  goetz
 dev_synch() now excludes udp clients and new imports

 Revision 8.1  2000/01/18 16:48:55  goetz
 tango_dev_import() now only called in dev_import if -DTANGO

 Revision 8.0  1999/12/28 14:18:35  goetz
 added TANGO support for TACO dev_xxx() calls via -DTANGO for C++

 Revision 7.10  1999/11/25 08:27:16  goetz
 replaced fprintf(stderr) with printf; made startup() C++ compatible

 Revision 7.9  1999/11/22 20:19:02  goetz
 removed const from gettransient() prototype

 Revision 7.8  1999/11/21 20:45:39  goetz
 included all M.Diehl's patches (major changes to gettransient() + main())

 Revision 7.7  1999/11/18 22:29:05  goetz
 fixed event multi-client bug, free client handle after event_listen, timeout in dev_synch()

 Revision 7.6  1999/11/01 19:05:17  goetz
 make local copy of user name returned by getpwuid() in sec_api.c

 Revision 7.5  1999/08/06 17:44:19  goetz
 removed usage of varargs for g++ on solaris and hpux

 * Revision 7.4  99/07/09  05:15:29  05:15:29  goetz (Andy Goetz)
 * added M.Diehl's patch to DevServerSig.c to exit() after calling unregister_server()
 * 
 Revision 7.3  1999/06/07 15:27:07  goetz
 fixed bug with multi-nethost reimport, device name stored with nethost

 Revision 7.2  1999/05/12 15:22:10  goetz
 changed dev_event_fire() to void; fixed bug in dev_event_unlisten()

 * Revision 7.1  99/05/11  15:59:57  15:59:57  goetz (Andy Goetz)
 * replace static declaration of event_client[] array by malloc()
 * 
 Revision 7.0  1999/04/26 07:30:48  goetz
 implemented user events (added event_api.c)

 Revision 6.15  1999/03/27 09:00:27  goetz
 asynch_cleanup() only checks for asynch clients; C++ returns output arguments if DS_WARNING

 * Revision 6.14  99/02/27  15:21:27  15:21:27  goetz (Andy Goetz)
 * fixed reimport bug; disabled reimport for rpc version 1
 * 
 Revision 6.13  1999/02/05 00:50:34  goetz
 fixed reimport socket+memory leak; ported to aCC on HPUX; suppressed rpc errors

 * Revision 6.12  99/01/18  21:39:11  21:39:11  goetz (Andy Goetz)
 * suppressed test for RPC_TIMEDOUT in svc_check()
 * 
 * Revision 6.11  99/01/14  23:15:01  23:15:01  goetz (Andy Goetz)
 * dbase timeout = 4 s; unregister ver. 1; svc_check() handles RPC_TIMEDOUT
 * 
 * Revision 6.10  98/12/22  10:29:53  10:29:53  taurel (E.Taurel)
 * Added support for Solaris and HP C++ native compiler. Remove K_R C support
 * 
 * Revision 6.9  98/12/15  14:05:14  14:05:14  goetz (Andy Goetz)
 * unregister RPC; use pgm no. 1 as flag; dev_pending() checks timeout; extern C prototype
 * 
 * Revision 6.8  98/11/20  14:35:42  14:35:42  goetz (Andy Goetz)
 * unregister API_VERSION (4) correctly now + check return value of svc_register()
 * 
 * Revision 6.7  98/11/20  10:35:08  10:35:08  goetz (Andy Goetz)
 * improved gettrans_ut(); multi-config_flags=0; removde static client_data; fixed linux select() bug; only register pgm # 0,4,5; portmap address == 111
 * 
 * Revision 6.6  98/09/24  09:25:15  09:25:15  goetz (Andy Goetz)
 * fixed stupid type which reintroduced asynchronous memory leak
 * 
 * Revision 6.5  98/08/26  16:20:39  16:20:39  goetz (Andy Goetz)
 * added dynamic NETHOST allocation, fixed bug in db_setup_multi, ported to Irix 6.4
 * 
 * Revision 6.4  98/04/29  10:21:27  10:21:27  goetz (Andy Goetz)
 * asynchronous - memory leak fixed, OS9 bug fixed, stale handle cleanup, pending=300
 * 
 * Revision 6.3  97/12/10  15:28:38  15:28:38  meyer (J.Meyer)
 * Changed select() timeout in rpc_check_host() to 200ms
 * 
 * Revision 6.2  97/12/01  12:59:21  12:59:21  goetz (Andy Goetz)
 * intermediate checkin before opening 5.14.1 branch
 * 
 * Revision 6.1  97/11/20  21:49:31  21:49:31  goetz (Andy Goetz)
 * main() continue even if cannot register asynch rpc
 * 
 * Revision 6.0  97/11/03  17:22:32  17:22:32  goetz (Andy Goetz)
 * merged synchronous and asynchronous branches
 * 
 Revision 5.15  1997/10/24 15:27:07  klotz
 ported to Windows NT 4.0 + NMSDK 5.0 ; tested with TextTalkds ; new DevServerMain.c

 * Revision 5.13  97/09/03  17:34:37  17:34:37  meyer (J.Meyer)
 * Added casting for ucc++ in DevServerSig.c
 * 
 * Revision 5.12  97/09/03  15:54:14  15:54:14  meyer (J.Meyer)
 * Changed OS9 signal handling from intercept() to signal().
 * 
 * Revision 5.11  97/08/05  14:19:33  14:19:33  goetz (Andy Goetz)
 * ported to VxWork 5.2, pgm. no. based on CRC, -m suppressed, -ve errors bug fix
 * 
 * Revision 5.10  97/03/14  17:26:13  17:26:13  dserver ()
 * fixed bug with open file handles after free
 * 
 * Revision 5.8.1.4  97/03/14  17:24:53  17:24:53  goetz (Andy Goetz)
 * fixed bug with open file handles after free
 * 
 * Revision 5.9  97/03/13  11:17:16  11:17:16  klotz (W.D. Klotz)
 * first part of NT4.0 port
 * 
 * Revision 5.8  97/01/31  12:06:25  12:06:25  goetz (Andy Goetz)
 * dev_protocol(), Linux+security bug fixes, dev_import() ignores timeout on NULL procedure, svc_run for OS9
 * 
 * Revision 5.7  97/01/14  08:57:21  08:57:21  goetz (Andy Goetz)
 * added stateless to dev_free(),_raw(),_asyn(),_protocol(),_timeout()
 * 
 * Revision 5.6  97/01/12  18:33:41  18:33:41  goetz (Andy Goetz)
 * dev_import() now stateless i.e. import works even if server is not running
 * 
 * Revision 5.5  97/01/08  17:02:31  17:02:31  goetz (Andy Goetz)
 * Ported to Linux (2.0.9)
 * 
 * Revision 5.4  96/11/27  09:58:16  09:58:16  goetz (Andy Goetz)
 * removed rpc admin interface from DevServerMain.c; putenv() suppressed for _UCC
 * 
 * Revision 5.3  96/11/26  16:24:42  16:24:42  goetz (Andy Goetz)
 * fixed bug (i_nethost=0) in dev_import_local()
 * 
 * Revision 5.2  96/11/26  13:31:06  13:31:06  goetz (Andy Goetz)
 * OICDevice (C++ wrapper for OIC) added + fixed multi-nethost memory leaks
 * 
 * Revision 5.1  96/11/04  10:55:24  10:55:24  goetz (Andy Goetz)
 * added multi nethost support
 * 

 Copyleft (c) 1990 by European Synchrotron Radiation Facility, 
                      Grenoble, France

********************************************************************-*/

#ifndef _DevSignal_h
#define _DevSignal_h

/*
 *  standart header string to use "what" or "ident".
 */
#ifdef _IDENT
static char DevSignalh[] =
"@(#)$Header: DevSignal.h,v 8.14 2000/06/22 14:50:41 goetz Rel $";
#endif /* _IDENT */


#ifdef OSK
/*
 * Get OS9 signal definition file
 */

#include <signal.h> 

/*
 * Define additional makros for OS9 
 */

#define SIGALRM 5
#define SIGTERM 6
#define SIGUSR1 7
#define SIGUSR2 8

/*
 * private OS9 signal definitions
 */
#define TIMEOUT_PLC	256
#define RECEIVE_PLC 	257
/*
 * allow also to use signal numbers up
 * to 512 without signal definition in this file.
 */
#define NUSIG	513   /* NSIG is 1 greater than the highest defined signal */
#else

#ifdef _OSK

/*
 * Ultra-C compatible code 
 *
 * in this case the signals SIGALRM, SIGTERM, SIGUSR1 and SIGUSR2 are defined 
 * as part of the POSIX set of signals and do/must not need to be defined
 * separately
 *
 * andy 8sep94
 */

#include <signal.h>

#define NUSIG	513   /* NUSIG is 1 greater than the highest defined signal */
                      /* use the same definition as used for Unibridge OSK  */

#endif /* _OSK */

/*
 * Get normal UNIX system signal definition file.
 * Set the number of available signals for UNIX systems
 */

#ifdef __hpux
#include <sys/signal.h>

#define NUSIG	_NSIG
#endif /*__hpux */

#if defined (solaris) || (sun) || (defined irix)
#include <signal.h>
#define NUSIG	NSIG
#endif /* solaris */

#ifdef lynx
#include <signal.h>

#define SIG_ERR -1 /* SIG_ERR is NOT defined in signal.h, */
		   /* contrary to what is said in the doc */

#define NUSIG	_NSIG
#endif /* lynx */
#endif /* OSK */


#if defined(_NT)
#include <signal.h>
#define NUSIG   NSIG
#endif  /* _NT */

#ifdef linux
#include <signal.h>
#define NUSIG   _NSIG
#endif /* linux */

#ifdef vxworks
#include <signal.h>
#define NUSIG	_NSIGS
#endif /* vxworks */

#endif /* _DevSignal_h */

